//инициализация библиотек, использующихся в программе
#include "Wire.h"     //библиотеки протокола I2C 
#include "I2Cdev.h"
#include "MPU6050.h"  //акселерометр-гироскоп MPU6050
#include "math.h"     //стандартная библиотека для вычислений

/**********************************************************************************************************************/
// Коэфициенты ПИД-алгоритма
// Меняя данные коэффициенты, можно настраивать поведение робота, добиваясь более стабильнй работы:
// Чтобы лучше понять влияние каждого из коэффициентов, надо рабобраться с формулой вычисления мощности двигателей:
// motorPower = Kp*(error) + Ki*(errorSum)*sampleTime - Kd*(currentAngle-prevAngle)/sampleTime

#define Kp  25      // Пропорциональный коэффициент.
                    // Чем выше значение этого коэффициента, тем сильнее моторы реагируют на изменение угла наклона с момента последнего опроса гироскопа
                    // При слишком малых значениях Kp робот не будет успевать восстановить равновесие при падении, 
                    // при слишком больших - моторы будут стартовать слишком резко, что уменьшит устойчивость


#define Ki  80      // Интергирующий коэффициент. 
                    // Чем выше значение этого коэффициента, тем сильнее моторы реагируют на накопленную ошибку отклонения от вертикали
                    // Увеличение этого коэффициента помогает бороться со статистической ошибкой, но увеличивает перерегулирование 

#define Kd  0.20    // Дифференциальный коэффициент
                    // Увеличение этого коэффициента позволяет бороться с перерегулированием, но заодно увеличивает влияние шумов

#define targetAngle 2.87      // Вертикаль робота (в градусах)
                              // Позволяет компенсировать постоянную ошибку определения вертикали и несовершенство геометрии робота
                              // Увеличивайте или уменьшайте это значение в зависимости от того, в какую сторону склонен смещаться робот со временем 

#define sampleTime  0.005  // Время между измерениями угла наклона (в сек). Этот параметр менять не следует.
 
/**********************************************************************************************************************/

MPU6050 mpu;  //объявление объекта MPU6050

//объявление переменных
int16_t accY, accZ, gyroX;  
volatile int motorPower, gyroRate;
volatile float accAngle, gyroAngle, currentAngle, prevAngle=0, error, prevError=0, errorSum=0;
volatile byte count=0;    //счетчик прерывания

void setMotors(int leftMotorSpeed, int rightMotorSpeed) {  //функция управления двигателями 
  if(leftMotorSpeed >= 0) {
    analogWrite(3, leftMotorSpeed);
    digitalWrite(5, LOW);
    digitalWrite(4, HIGH);
  }
  else {
    analogWrite(3, abs(leftMotorSpeed));
    digitalWrite(5, HIGH);
    digitalWrite(4, LOW);
  }
  if(rightMotorSpeed >= 0) {
    analogWrite(6, rightMotorSpeed);
    digitalWrite(8, LOW);
    digitalWrite(7, HIGH);
  }
  else {
    analogWrite(6,  abs(rightMotorSpeed));
    digitalWrite(8, HIGH);
    digitalWrite(7, LOW);
  }
}
 
void init_PID() {  //объявление прерывания
  cli();          // отключение глобальных прерываний
  TCCR1A = 0;     // установить регистр TCCR1A на 0
  TCCR1B = 0;     // то же для TCCR1B   
  // установка таймера на 5 мс
  OCR1A = 9999;    
  // включить режим CTC
  TCCR1B |= (1 << WGM12);
  // Установка бита CS11 для предписания 8
  TCCR1B |= (1 << CS11);
  // включение сравнения прерывания таймера
  TIMSK1 |= (1 << OCIE1A);
  sei();          // включение глобальных прерываний
}

void setup() {
  // установка двигателей и ШИМ в режим вывода
  pinMode(3, OUTPUT);
  pinMode(4, OUTPUT);
  pinMode(5, OUTPUT);
  pinMode(6, OUTPUT);
  pinMode(7, OUTPUT);
  pinMode(8, OUTPUT);
  pinMode(13, OUTPUT);
  mpu.initialize();  //включение гироскопа
      mpu.setXGyroOffset(200); //установка начальных данных гироскопа и акселерометра
      mpu.setYGyroOffset(100);
      mpu.setZAccelOffset(1788);

  init_PID();  //разрешение прерывания
}

void loop() 
{
  // получение данных с гироскопа и акселерометра
  accY = mpu.getAccelerationY();
  accZ = mpu.getAccelerationZ();  
  gyroX = mpu.getRotationX();
  // установка и ограничивание мощности моторов
  motorPower = constrain(motorPower, -255, 255);
  setMotors(motorPower, motorPower);
}

ISR(TIMER1_COMPA_vect)  //запуск прерывания каждые 5 миллисекунд
{
  // вычисление угла наклона
  accAngle = atan2(accY, accZ)*RAD_TO_DEG;
  gyroRate = map(gyroX, -32768, 32767, -250, 250);
  gyroAngle = (float)gyroRate*sampleTime;  
  currentAngle = 0.9934*(prevAngle + gyroAngle) + 0.0066*(accAngle);
  
  error = currentAngle - targetAngle;
  errorSum = errorSum + error;  
  errorSum = constrain(errorSum, -300, 300);

  // Вычисление мощности двигателей с учетом ПИД коэффициентов
    motorPower = Kp*(error) + Ki*(errorSum)*sampleTime - Kd*(currentAngle-prevAngle)/sampleTime;
 
  prevAngle = currentAngle;
  count++;
  if(count == 200)  {
    count = 0;
    digitalWrite(13, !digitalRead(13));
  }
}
